"""
Programm zum Durchführen von SQL-Injections

license information: static/LICENSE.txt

Konvertieren in eine ausführbare Datei
Windows:
pyinstaller -F -n sql-injection-dummy-windows --add-data "templates;templates" --add-data "static;static"  sql-injection-dummy.py
Ubuntu:
pyinstaller -F -n sql-injection-dummy-ubuntu --add-data "templates:templates" --add-data "static:static"  sql-injection-dummy.py
Mac:
pyinstaller -F -n sql-injection-dummy-mac --add-data "templates:templates" --add-data "static:static"  sql-injection-dummy.py
"""

import json
import os
import platform
import re
import socket
import sqlite3

import sys

from flask import Flask, render_template, request

app = Flask(__name__)

LEVELS = [
    {
        "massnahmen": "Keine besonderen Sicherheitsmaßnahmen",
        "einleitung": "",
        "aufgaben": [
            {
                "text": """Geben Sie im Mail-Adressen-Feld kontakt@peter.de und im Passwort-Feld 1234 ein. Drücken Sie dann auf Anmelden. <br><br>
                    Unter dem Anmeldefeld wird jetzt ein blauer Kasten angezeigt. Dieser enthält die SQL-Abfrage, die der Webserver an den DB-Server schickt. Der Webserver setzt in diesem Level die eingegebenen Texte für die Mail-Adresse und das Passwort direkt ein. <br>
                    Die SQL-Abfrage soll überprüfen, ob ein Nutzer mit den eingegebenen Werten für die Mail-Adresse und das Passwort in der Datenbank existiert. Erhält der Webserver vom DB-Server ein Ergebnis, erlaubt er die Anmeldung. <br><br> Das Vorgehen hat momentan ein großes Problem: Auch SQL-Code, den wir in die Eingabefelder eingeben, wird direkt in die Abfrage eingesetzt.
                    """,
                "tipps": [],
            },
            {
                "text": "Finden Sie eine Eingabe für das Passwort-Feld, sodass der Webserver zwei Abfragen auf einmal an die Datenbank sendet. Die zweite Abfrage soll nach dem Nutzer mit der Mail-Adresse 'peter.karlsson@gmx.de' fragen. <br><br>Sobald Sie sich erfolgreich angemeldet haben, wird Level 2 freigeschaltet. Wechseln Sie dann zu Level 2.",
                "tipps": [
                    """Die folgende Abfrage gibt den Nutzer mit der Mail-Adresse 'peter.karlsson@gmx.de' zurück. <br>
                        <div class='sql'>SELECT * FROM Nutzer WHERE MailAdresse='peter.karlsson@gmx.de';</div> <br> Sorgen Sie durch eine geschickte Eingabe ins Passwort-Feld dafür, dass diese Abfrage Teil von gültigem SQL-Code wird, der an den DB-Server übermittelt wird.""",
                    "SQL-Befehle beendet man mit einem Semikolon (;)",
                    """Insgesamt soll der DB-Server folgenden SQL-Code erhalten:
                        <br> <div class='sql'>SELECT * FROM Nutzer WHERE MailAdresse = '' AND Passwort = ''; SELECT * FROM Nutzer WHERE MailAdresse='peter.karlsson@gmx.de';</div> <br> Sorgen Sie durch eine geschickte Eingabe ins Passwort-Feld dafür, dass genau dieser SQL-Code an den DB-Server übermittelt wird.""",
                ],
            },
        ],
    },
    {
        "massnahmen": "Neue Sicherheitsmaßnahme: Ab diesem Level können nicht mehr mehrere SQL-Abfragen auf einmal ausgeführt werden. ",
        "einleitung": "",
        "aufgaben": [
            {
                "text": """Testen Sie Ihren Angriff aus Level 1. Beschreiben Sie, was passiert. <br><br><br> Welche Datensätze eine SQL-Abfrage liefert, wird durch die Bedingungen nach dem WHERE festgelegt. Die folgende Abfrage liefert z.B. nur alle Nutzer mit Mail-Adresse 'peter.karlsson@gmx.de' und Passwort '1234': <br>
                    <div class='sql'>SELECT * FROM Nutzer WHERE MailAdresse = 'peter.karlsson@gmx.de' AND Passwort='1234';</div> <br>
                    Da wir das korrekte Passwort aber dummerweise nicht kennen, wollen wir einen Trick anwenden. Durch eine geschickte Eingabe ins Passwort-Feld, können wir dafür sorgen, dass die gesamte WHERE-Bedingung immer wahr ist und wir somit angemeldet werden.""",
                "tipps": [],
            },
            {
                "text": "Finden Sie eine Eingabe für das Passwort-Feld, sodass eine einzige SQL-Abfrage an den DB-Server übermittelt wird, deren WHERE-Bedingung immer wahr ist.",
                "tipps": [
                    "Eine Bedingung die immer wahr ist, wäre z.B. 1 = 1 oder 'hi' = 'hi'. Eine Bedingung dieser Art sollte Teil unserer Gesamtabfrage werden.",
                    "Das logische ODER (in SQL: OR) ist immer wahr, wenn mindestens eine der verknüpften Bedingungen wahr ist.",
                    "In SQL werden Bedingungen, die durch AND verknüpft sind, zuerst ausgewertet und danach erst solche, die mit OR verknüpft sind. <br> A AND B OR C ist gleichwertig zu (A AND B) OR C",
                    """Der folgende SQL-Befehl gibt alle Nutzer aus und sorgt für eine Anmeldung als erster gespeicherter Nutzer: <br>
                        <div class='sql'>SELECT * FROM Nutzer WHERE MailAdresse = '' AND Passwort = '' OR '1' = '1'</div> <br>
                        Finden Sie eine geschickte Eingabe für das Passwort-Feld, sodass dieser Befehl abgeschickt wird.""",
                ],
            },
            {
                "text": "(Zusatzaufgabe) Finden Sie eine Eingabe für das Passwort-Feld, sodass Sie als anton.goethe@gmx.de angemeldet werden.",
                "tipps": [
                    "In SQL werden Bedingungen, die durch AND verknüpft sind, zuerst ausgewertet und danach erst solche, die mit OR verknüpft sind. <br> A AND B OR C ist gleichwertig zu (A AND B) OR C",
                    "Bei Ihrer Abfrage aus der vorherigen Aufgabe haben Sie eine Bedingung genutzt, die immer wahr ist. Nutzen Sie nun eine Bedingung, die nur für den gewünschten Datensatz erfüllt ist",
                ],
            },
            {
                "text": "Bewerten Sie, wie gut die neue Sicherheitsmaßnahme gegen SQL-Injections hilft. Notieren Sie insbesondere, gegen welche Art von Attacken sie hilfreich ist und gegen welche sie nicht hilft.",
                "tipps": [
                    "In SQL gibt es außer SELECT-Abfragen auch Befehle, die die Datenbank manipulieren, also z.B. vorhandene Daten ändern oder löschen."
                ],
            },
        ],
    },
    {
        "massnahmen": "Neue Sicherheitsmaßnahme: Ab diesem Level wird das Schlüsselwort OR aus deiner Eingabe gefiltert.",
        "einleitung": "",
        "aufgaben": [
            {
                "text": """Testen Sie Ihren Angriff aus Level 2. Beschreiben Sie, was passiert. <br><br><br>
                    In SQL kann man mit den zwei Zeichen -- einen Kommentar beginnen. Der folgende Befehl ist auskommentiert und wird somit nicht ausgeführt: <br>
                    <div class='sql'>-- SELECT * FROM Nutzer;</div>""",
                "tipps": [],
            },
            {
                "text": "Finden Sie eine Eingabe für die Mail-Adresse, sodass die Abfrage an die Datenbank nur die Mail-Adresse abfragt und der Passwort-Teil der Abfrage auskommentiert wird. Melden Sie sich als Nutzer 'peter.karlsson@gmx.de' an. Sobald Ihre Injection erfolgreich war, wird Level 4 freigeschaltet.",
                "tipps": [],
            },
            {
                "text": "Bewerten Sie, wie gut die neue Sicherheitsmaßnahme gegen SQL-Injections hilft. Notieren Sie insbesondere, gegen welche Art von Attacken sie hilfreich ist und gegen welche sie nicht hilft.",
                "tipps": [],
            },
        ],
    },
    {
        "massnahmen": "Neue Sicherheitsmaßnahme: Ab diesem Level werden SQL-Parameter genutzt. Die Abfrage und die Parameter werden dabei getrennt an den DB-Server übermittelt. Code in den Parametern wird nicht ausgeführt. Die Parameter werden als reine Texte, Zahlen, etc. behandelt.",
        "einleitung": "",
        "aufgaben": [
            {
                "text": "Testen Sie Ihre Abfragen aus den vorherigen Levels. Beschreiben Sie, was passiert.",
                "tipps": [],
            },
            {
                "text": "Bewerten Sie, wie gut die neue Sicherheitsmaßnahme gegen SQL-Injections hilft. Notieren Sie insbesondere, gegen welche Art von Attacken sie hilfreich ist und gegen welche sie nicht hilft.",
                "tipps": [],
            },
        ],
    },
    {
        "massnahmen": "Keine besonderen Sicherheitsmaßnahmen mehr aktiviert.",
        "einleitung": "Das Schema der Tabelle Nutzer sieht folgendermaßen aus: <br> Nutzer(<u>MailAdresse</u>, Name, Vorname, Guthaben, Passwort)",
        "aufgaben": [
            {
                "text": "Ändern Sie das Passwort für alle Nutzer mit dem Namen Peter Karlsson auf '1234'. Verwenden Sie dazu den Update-Befehl. <br> Syntax: <div class='sql'>UPDATE Tabellenname SET Attributname = neuerWert WHERE Bedingung;</div> <br> Beispiel: <div class='sql'>UPDATE Nutzer SET Name='Meier' WHERE Name = 'Maier';</div>",
                "tipps": ["Schauen Sie sich das Schema der Tabelle Nutzer oben genau an."],
            },
            {
                "text": "Melden Sie sich mit der Mail-Adresse 'peter.karlsson@gmx.de' und dem neu gesetzten Passwort '1234' an. Überprüfen Sie, dass das nun möglich ist.",
                "tipps": [],
            },
            {
                "text": "Ändern Sie das Guthaben von 'anton.schmidt@gmx.de' auf den höchsten Wert, der in der DB existiert. Diesen Wert können Sie mit dem Befehl <br><div class='sql'>(SELECT max(Guthaben) FROM Nutzer)</div> <br> bestimmen. Diesen SQL-Befehl können Sie einfach beim Update für den neuen Wert einsetzen. Melden Sie sich vorher und nachher an. Überprüfen Sie, dass die Änderung korrekt ausgeführt wurde.",
                "tipps": [],
            },
            {
                "text": "Ändern Sie das Guthaben von 'anton.schmidt@gmx.de' auf den niedrigsten Wert, der in der DB existiert. Diesen Wert können Sie mit dem Befehl <br><div class='sql'>(SELECT min(Guthaben) FROM Nutzer)</div> <br> bestimmen. Melden Sie sich vorher und nachher an. Überprüfen Sie, dass die Änderung korrekt ausgeführt wurde.",
                "tipps": [],
            },
            {
                "text": "Sorgen Sie mit einem einzigen SQL-Befehl dafür, dass niemand mehr Schulden hat. Notieren Sie diesen Befehl.",
                "tipps": [],
            },
        ],
    },
    {
        "massnahmen": "Keine besonderen Sicherheitsmaßnahmen aktiviert.",
        "einleitung": "Das Schema der Tabelle Nutzer sieht folgendermaßen aus: <br> Nutzer(<u>MailAdresse</u>, Name, Vorname, Guthaben, Passwort)",
        "aufgaben": [
            {
                "text": "Legen Sie einen neuen Account mit Ihrer Mail-Adresse an. Verwenden Sie dazu den Insert-Befehl. <br> Syntax: <div class='sql'>INSERT INTO Tabellenname VALUES (WertFuerAttribut1, WertFuerAttribut2, ...);</div> <br> Beispiel: <div class='sql'>INSERT INTO Nutzer VALUES ('hansi@gmx.de', 'Müller', 'Hansi', 0.00, '1234');</div>",
                "tipps": [],
            },
            {
                "text": "Melden Sie sich mit Ihrer Mail-Adresse und dem festgelegten Passwort an und überprüfen Sie, dass das nun möglich ist und das Guthaben korrekt festgelegt wurde.",
                "tipps": [],
            },
            {
                "text": "Löschen Sie den Account von 'peter.karlsson@gmx.de'. Verwenden Sie dazu den Delete-Befehl. <br> Syntax: <div class='sql'>DELETE FROM Tabellenname WHERE Bedingung;</div> <br> Beispiel: <div class='sql'>DELETE FROM Nutzer WHERE Guthaben > 0;</div>",
                "tipps": [],
            },
            {
                "text": "Ihnen tut Peter Karlsson nun doch leid. Legen Sie seinen Account mit einem ordentlichen Guthaben und dem Passwort '1234' wieder an. Verwenden Sie den Insert-Befehl",
                "tipps": [],
            },
            {
                "text": "Melden Sie sich mit der Mail-Adresse 'peter.karlsson@gmx.de' und dem Passwort '1234' an und überprüfen Sie, dass das nun wieder möglich ist.",
                "tipps": [],
            },
        ],
    },
    {
        "massnahmen": "Keine besonderen Sicherheitsmaßnahmen mehr aktiviert.",
        "einleitung": "Das Schema der Tabelle Nutzer sieht folgendermaßen aus: <br> Nutzer(<u>MailAdresse</u>, Name, Vorname, Guthaben, Passwort)",
        "aufgaben": [
            {
                "text": "Löschen Sie die Tabelle Nutzer. Recherchieren Sie im Internet, mit welchem Befehl das geht.",
                "tipps": [],
            },
            {
                "text": "Versuchen Sie sich mit der Mail-Adresse 'peter.karlsson@gmx.de' und dem Passwort '1234' anzumelden. Überprüfen Sie, dass das nun nicht mehr möglich ist.",
                "tipps": [],
            },
            {
                "text": "Klicken Sie auf den Button, um den ursprünglichen Zustand der Datenbank wiederherzustellen. Alle Ihre Änderungen werden dadurch rückgängig gemacht! <input type='submit' value='DB wiederherstellen' onclick='db_reset()'>",
                "tipps": [],
            },
            {
                "text": """Erstellen Sie eine Präsentation, die die folgenden Punkte betrachtet:
                <ul>
                    <li> Was ist eine SQL-Injection? </li>
                    <li> Nennen Sie 3 Sicherheitsmaßnahmen zum Verhindern von SQL-Injections und bewerten Sie deren Effektivität. </li>
                    <li> Darf man auf einer beliebigen Seite eine SQL-Injection durchführen? </li>
                </ul>
                Ergänzen Sie Ihre Präsentation gerne um zusätzliche Bilder und Informationen aus dem Internet. Vergessen Sie aber nicht, jeweils die genaue Quelle anzugeben.""",
                "tipps": [],
            },
        ],
    },
]


@app.route("/")
def start():
    """
    Initialer Aufbau der Seite"""

    return render_template(
        "login.html",
        levels=LEVELS,
    )


@app.route("/query", methods=["POST"])
def query():

    daten = request.get_json()
    if daten is not None:
        db = sqlite3.connect(db_pfad)
        level = daten["level_nr"]
        mail = daten["mail"]
        passwort = daten["pw"]
        c = db.cursor()
        sql_befehl = """SELECT * FROM Nutzer 
        WHERE MailAdresse = '{}' AND Passwort='{}';""".format(
            mail, passwort
        )
        try:
            print(sql_befehl)
            if level in [1, 5, 6, 7]:
                # Workaround: Mehrere Befehle zugelassen
                # In SQLite grundsätzlich nicht erlaubt
                befehle = sql_befehl.split(";")
                for befehl in befehle:
                    befehl = befehl.strip()
                    if len(befehl) > 1 and befehl[:2] != "--":
                        c.execute(befehl)
            elif level == 2:
                # Nur noch ein Befehl erlaubt
                c.execute(sql_befehl)
            elif level == 3:
                # Filter OR mit RegEx, Groß-Klein-Schreibung ignorieren
                sql_befehl = re.sub(" OR ", " ", sql_befehl, flags=re.IGNORECASE)
                sql_befehl = re.sub("'OR ", "' ", sql_befehl, flags=re.IGNORECASE)
                sql_befehl = re.sub(" OR'", " '", sql_befehl, flags=re.IGNORECASE)
                sql_befehl = re.sub("'OR'", "''", sql_befehl, flags=re.IGNORECASE)
                c.execute(sql_befehl)
            elif level == 4:
                # SQL-Parameter genutzt
                c.execute(
                    """SELECT * FROM Nutzer WHERE MailAdresse = ? AND Passwort = ?""",
                    (mail, passwort),
                )
                sql_befehl = "SELECT * FROM Nutzer WHERE MailAdresse = ? AND Passwort = ? <br><br>Parameter: <br> (1) {}<br>(2) {}".format(
                    mail, passwort
                )
        except Exception as e:
            fehlermeldung = repr(e)
            db.close()
            return (
                json.dumps(
                    {
                        "success": True,
                        "angemeldet": False,
                        "status": "Anmeldung fehlgeschlagen<br>" + fehlermeldung,
                        "sql": sql_befehl,
                    }
                ),
                200,
                {"ContentType": "application/json"},
            )

        res = c.fetchall()
        db.commit()
        db.close()

        if len(res) > 0:
            return (
                json.dumps(
                    {
                        "success": True,
                        "angemeldet": True,
                        "status": "Anmeldung erfolgreich<br><br>Angemeldet als: {}<br>Guthaben: {}€".format(
                            res[0][0], res[0][3]
                        ),
                        "sql": sql_befehl,
                    }
                ),
                200,
                {"ContentType": "application/json"},
            )
        else:
            return (
                json.dumps(
                    {
                        "success": True,
                        "angemeldet": False,
                        "status": "Anmeldung fehlgeschlagen",
                        "sql": sql_befehl,
                    }
                ),
                200,
                {"ContentType": "application/json"},
            )

    return (
        json.dumps({"success": False, "angemeldet": False, "status": "", "sql": ""}),
        200,
        {"ContentType": "application/json"},
    )


@app.route("/db-reset", methods=["GET"])
def db_reset():
    db_erstellen()
    return (json.dumps({"success": True}), 200, {"ContentType": "application/json"})


def db_erstellen():
    print("DB wird neu erstellt")
    db = sqlite3.connect(db_pfad)
    c = db.cursor()
    if getattr(sys, "frozen", False):
        # Lesen in der PyInstaller-Exe
        with open(
            os.path.join(sys._MEIPASS, "static/nutzerdaten.sql"),
            encoding="utf-8",
            mode="r",
        ) as f:
            c.executescript(f.read())
    else:
        with open("static/nutzerdaten.sql", encoding="utf-8", mode="r") as f:
            c.executescript(f.read())

    db.commit()
    db.close()


def port_wird_genutzt(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


if __name__ == "__main__":
    arbeitsverzeichnis = os.getcwd()
    betriebssystem = platform.system().lower()
    if betriebssystem == "darwin":
        # Darwin -> MacOS, unter Mac funktioniert getcwd() in der .app nicht mehr
        if getattr(sys, "frozen", False):
            arbeitsverzeichnis = os.path.dirname(sys.executable)
    db_pfad = os.path.join(arbeitsverzeichnis, "nutzerdaten.db")
    if not os.path.exists(db_pfad):
        db_erstellen()
    
    # ersten freien Port auswählen
    for port in range(1024, 36000):

        if not port_wird_genutzt(port):
            print("Freier Port, URL: http://127.0.0.1:{}/".format(port))
            app.run(port=port)
            break
        else:
            print("Port {} belegt".format(port))
